from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID, uuid4

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, declared_attr, mapped_column, relationship
from uuid_extensions import uuid7str

from pyidentity.contrib.sqlalchemy.base.abstract import (
    AbstractIdentityRole,
    AbstractIdentityRoleClaim,
    AbstractIdentityUser,
    AbstractIdentityUserClaim,
    AbstractIdentityUserLogin,
    AbstractIdentityUserRole,
    AbstractIdentityUserToken,
)
from pyidentity.contrib.sqlalchemy.base.model import Model

if TYPE_CHECKING:
    from datetime import datetime

    from pyidentity.models import TKey

__all__ = (
    "IdentityRole",
    "IdentityRoleClaim",
    "IdentityUser",
    "IdentityUserClaim",
    "IdentityUserLogin",
    "IdentityUserRole",
    "IdentityUserToken",
    "Model",
)


class IdentityUser(AbstractIdentityUser):
    """The default implementation of AbstractIdentityUser which uses a string as a primary key."""

    __personal_data__ = (
        "id",
        "username",
        "email",
        "email_confirmed",
        "phone_number",
        "phone_number_confirmed",
        "two_factor_enabled",
    )

    if TYPE_CHECKING:
        access_failed_count: int
        concurrency_stamp: str | UUID | None
        email: str | None
        email_confirmed: bool
        id: TKey
        lockout_enabled: bool
        lockout_end: datetime | None
        normalized_email: str | None
        normalized_username: str | None
        password_hash: str | None
        phone_number: str | None
        phone_number_confirmed: bool
        security_stamp: str | UUID | None
        two_factor_enabled: bool
        username: str | None

    id: Mapped[str] = mapped_column(sa.String(450))

    @declared_attr
    def roles(self) -> Mapped[list[IdentityRole]]:
        return relationship(
            "IdentityRole",
            back_populates="users",
            secondary="pydentity_user_roles",
            cascade="all, delete",
        )

    @declared_attr
    def claims(self) -> Mapped[list[IdentityUserClaim]]:
        return relationship("IdentityUserClaim", back_populates="user", cascade="all, delete")

    @declared_attr
    def logins(self) -> Mapped[list[IdentityUserLogin]]:
        return relationship("IdentityUserLogin", back_populates="user", cascade="all, delete")

    @declared_attr
    def tokens(self) -> Mapped[list[IdentityUserToken]]:
        return relationship("IdentityUserToken", back_populates="user", cascade="all, delete")

    def __init__(self, email: str, username: str | None = None, **kwargs) -> None:
        super().__init__(
            id=uuid7str(),
            email=email,
            username=username,
            security_stamp=str(uuid4()),
            **kwargs,
        )


class IdentityRole(AbstractIdentityRole):
    """The default implementation of AbstractIdentityRole which uses a string as the primary key."""

    if TYPE_CHECKING:
        concurrency_stamp: str | UUID | None
        id: TKey
        name: str | None
        normalized_name: str | None

    id: Mapped[str] = mapped_column(sa.String(450))

    @declared_attr
    def users(self) -> Mapped[list[IdentityUser]]:
        return relationship("IdentityUser", back_populates="roles", secondary="pydentity_user_roles")

    @declared_attr
    def claims(self) -> Mapped[list[IdentityRoleClaim]]:
        return relationship("IdentityRoleClaim", back_populates="role", cascade="all, delete")

    def __init__(self, name: str, **kwargs) -> None:
        super().__init__(id=uuid7str(), name=name, **kwargs)


class IdentityUserRole(AbstractIdentityUserRole):
    """Represents the link between a user and a role."""


class IdentityUserClaim(AbstractIdentityUserClaim):
    """Represents a claim that a user possesses."""

    @declared_attr
    def user(self) -> Mapped[IdentityUser]:
        return relationship("IdentityUser", back_populates="claims")


class IdentityUserLogin(AbstractIdentityUserLogin):
    """Represents a login and its associated provider for a user."""

    @declared_attr
    def user(self) -> Mapped[IdentityUser]:
        return relationship("IdentityUser", back_populates="logins")


class IdentityUserToken(AbstractIdentityUserToken):
    """Represents an authentication token for a user."""

    @declared_attr
    def user(self) -> Mapped[IdentityUser]:
        return relationship("IdentityUser", back_populates="tokens")


class IdentityRoleClaim(AbstractIdentityRoleClaim):
    """Represents a claim that is granted to all users within a role."""

    @declared_attr
    def role(self) -> Mapped[IdentityRole]:
        return relationship("IdentityRole", back_populates="claims")
